"""
Geonomics Help (:mod:`geonomics.help`)
======================================

.. currentmodule:: geonomics.help

Memory Help
-----------

    get_obj_size    Get the full size of a complex object

Parameters Help
---------------

    plot_movement   Plot example movement tracks generated by the input
                    movement parameters


Background information
----------------------

Geonomics models can occupy a lot of memory, and also contain a large number of
parameters to be customized. The functions in this module aim to provide some
assistance in assessing Geonomics' behavior, so as to help users better plan,
introspect, and debug their models.

"""
from ._param_help import *
from ._memory_help import *
